import enum
from typing import List, Iterable

from pysh.syntaxnodes import SyntaxNode, ExpressionNode, ArgumentPartNode, ArgumentNode, ArgumentPartType, \
    SyntaxNodeVisitor


class OpCode(enum.Enum):
    NoOp = 0
    Concat = 1
    Substitute = 2
    SubstituteSingle = 3
    LoadBuffer = 4
    PushBuffer = 5
    ResetA = 6
    IncrementA = 7
    PushA = 8
    Call = 9
    BranchReturnValue = 10
    BranchBufferEmpty = 11
    JumpRelative = 12
    DefineFunction = 13


class Instruction(object):
    @property
    def opcode(self) -> OpCode:
        return OpCode.NoOp

    def accept(self, visitor: 'InstructionVisitor') -> None:
        pass


class ConcatInstruction(Instruction):
    def __init__(self, value: str) -> None:
        self.value = value

    @property
    def opcode(self) -> OpCode:
        return OpCode.Concat

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_concat(self)


class SubstituteInstruction(Instruction):
    def __init__(self, value: str) -> None:
        self.value = value

    @property
    def opcode(self) -> OpCode:
        return OpCode.Substitute

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_substitute(self)


class SubstituteSingleInstruction(Instruction):
    def __init__(self, value: str) -> None:
        self.value = value

    @property
    def opcode(self) -> OpCode:
        return OpCode.SubstituteSingle

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_substitute_single(self)


class LoadBufferInstruction(Instruction):
    def __init__(self, value: str) -> None:
        self.value = value

    @property
    def opcode(self) -> OpCode:
        return OpCode.LoadBuffer

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_load_buffer(self)


class PushBufferInstruction(Instruction):
    def opcode(self) -> OpCode:
        return OpCode.PushBuffer

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_push_buffer(self)


class ResetAInstruction(Instruction):
    def opcode(self) -> OpCode:
        return OpCode.PushBuffer

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_reset_a(self)


class IncrementAInstruction(Instruction):
    def opcode(self) -> OpCode:
        return OpCode.IncrementA

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_increment_a(self)


class PushAInstruction(Instruction):
    def opcode(self) -> OpCode:
        return OpCode.PushA

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_push_a(self)


class CallInstruction(Instruction):
    @property
    def opcode(self) -> OpCode:
        return OpCode.Call

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_call(self)


class BranchReturnValueInstruction(Instruction):
    def __init__(self, offset: int) -> None:
        self.offset = offset

    @property
    def opcode(self) -> OpCode:
        return OpCode.BranchReturnValue

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_branch_return_value(self)


class BranchBufferEmptyInstruction(Instruction):
    def __init__(self, offset: int) -> None:
        self.offset = offset

    @property
    def opcode(self) -> OpCode:
        return OpCode.BranchBufferEmpty

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_branch_buffer_empty(self)


class JumpRelativeInstruction(Instruction):
    def __init__(self, offset: int) -> None:
        self.offset = offset

    @property
    def opcode(self) -> OpCode:
        return OpCode.JumpRelative

    def accept(self, visitor: 'InstructionVisitor') -> None:
        visitor.visit_jump_relative(self)


class InstructionVisitor(object):
    def visit_concat(self, instruction: ConcatInstruction) -> None:
        pass

    def visit_substitute(self, instruction: SubstituteInstruction) -> None:
        pass

    def visit_substitute_single(self, instruction: SubstituteSingleInstruction) -> None:
        pass

    def visit_load_buffer(self, instruction: LoadBufferInstruction) -> None:
        pass

    def visit_push_buffer(self, instruction: PushBufferInstruction) -> None:
        pass

    def visit_reset_a(self, instruction: ResetAInstruction) -> None:
        pass

    def visit_increment_a(self, instruction: IncrementAInstruction) -> None:
        pass

    def visit_push_a(self, instruction: PushAInstruction) -> None:
        pass

    def visit_call(self, instruction: CallInstruction) -> None:
        pass

    def visit_branch_return_value(self, instruction: BranchReturnValueInstruction) -> None:
        pass

    def visit_branch_buffer_empty(self, instruction: BranchBufferEmptyInstruction) -> None:
        pass

    def visit_jump_relative(self, instruction: JumpRelativeInstruction) -> None:
        pass


class CodeGenVisitor(SyntaxNodeVisitor):
    def __init__(self) -> None:
        self.code: List[Instruction] = []

    def visit_argument_part_node(self, node: ArgumentPartNode) -> None:
        pass

    def visit_argument_node(self, node: ArgumentNode) -> None:
        pass

    def visit_expression_node(self, node: ExpressionNode) -> None:
        self.code.append(LoadBufferInstruction(""))
        self.code.append(ResetAInstruction())
        for arg_node in node.args:
            last_part_was_replacement = False
            for part_node in arg_node.parts:
                last_part_was_replacement = False
                if part_node.type == ArgumentPartType.CONSTANT:
                    self.code.append(ConcatInstruction(part_node.value))
                elif part_node.type == ArgumentPartType.REPLACEMENT:
                    self.code.append(SubstituteInstruction(part_node.value))
                    last_part_was_replacement = True
                elif part_node.type == ArgumentPartType.REPLACEMENT_SINGLE:
                    self.code.append(SubstituteSingleInstruction(part_node.value))
                else:
                    raise Exception("bug")

            if last_part_was_replacement:
                self.code.append(BranchBufferEmptyInstruction(2))
            self.code.append(PushBufferInstruction())
            self.code.append(IncrementAInstruction())

        self.code.append(CallInstruction())


class CodeGenerator(object):
    def __init__(self) -> None:
        pass

    def generate(self, syntax_nodes: Iterable[SyntaxNode]) -> List[Instruction]:
        visitor = CodeGenVisitor()
        for node in syntax_nodes:
            node.accept(visitor)
        return visitor.code



import enum
from typing import List, Match, Tuple, Pattern, Callable
import re


class TokenType(enum.Enum):
    UNKNOWN = 0
    WHITESPACE = 1
    EOS = 2
    SYMBOL = 3
    REPLACEMENT = 4
    IF = 5
    THEN = 6
    FI = 7
    QUOTES = 8


class Token(object):
    @property
    def type(self) -> TokenType:
        return TokenType.UNKNOWN


class SimpleToken(Token):
    def __init__(self, type: TokenType) -> None:
        self._type = type

    def __repr__(self) -> str:
        return '<SimpleToken type: {0}>'.format(self._type)

    @property
    def type(self) -> TokenType:
        return self._type


class SymbolToken(Token):
    def __init__(self) -> None:
        self.value = ''

    def __repr__(self) -> str:
        return '<SymbolToken value: {0}>'.format(self.value)

    @property
    def type(self) -> TokenType:
        return TokenType.SYMBOL


class ReplacementToken(Token):
    def __init__(self) -> None:
        self.value = ''

    def __repr__(self) -> str:
        return '<ReplacementToken value: {0}>'.format(self.value)

    @property
    def type(self) -> TokenType:
        return TokenType.REPLACEMENT


class Lexer(object):
    whitespace_pattern = re.compile(r'^[^\S\n]+')
    newline_pattern = re.compile(r'^\n+')
    semicolon_pattern = re.compile(r'^;')
    quotes_pattern = re.compile(r'^"')
    symbol_pattern = re.compile(r'^[a-zA-Z]+')
    simple_replacement_pattern = re.compile(r'^\$([a-zA-Z]+)')
    block_replacement_pattern = re.compile(r'^\${([a-zA-Z]+)}')

    def __init__(self) -> None:
        pass

    def lex_all(self, source: str) -> List[Token]:
        tokens = []
        while len(source) > 0:
            token, source = self.lex(source)
            tokens.append(token)
        return tokens

    def lex(self, source: str) -> Tuple[Token, str]:
        for pattern, lex_func in __class__.PATTERNS:
            match = pattern.match(source)
            if match is None:
                continue
            token = lex_func(self, match)
            substr = source[match.end(0):]
            return token, substr
        return Token(), ''

    def lex_whitespace(self, match: Match) -> Token:
        return SimpleToken(TokenType.WHITESPACE)

    def lex_end_of_statement(self, match: Match) -> Token:
        return SimpleToken(TokenType.EOS)

    def lex_quotes(self, match: Match) -> Token:
        return SimpleToken(TokenType.QUOTES)

    def lex_symbol(self, match: Match) -> SymbolToken:
        token = SymbolToken()
        token.value = match.group(0)
        return token

    def lex_replacement(self, match: Match) -> ReplacementToken:
        token = ReplacementToken()
        token.value = match.group(1)
        return token

    PATTERNS: List[Tuple[Pattern, Callable[[Match], Token]]] = [
        (whitespace_pattern, lex_whitespace),
        (newline_pattern, lex_end_of_statement),
        (semicolon_pattern, lex_end_of_statement),
        (quotes_pattern, lex_quotes),
        (symbol_pattern, lex_symbol),
        (simple_replacement_pattern, lex_replacement),
        (block_replacement_pattern, lex_replacement)
    ]



from typing import List, Iterable, Optional
from pysh.lexer import Token, TokenType, SymbolToken, ReplacementToken
from pysh.syntaxnodes import SyntaxNode, ArgumentNode, ArgumentPartNode, ArgumentPartType, ExpressionNode


class ParseError(Exception):
    def __init__(self, *args):
        super(__class__, self).__init__(*args)


class StateTickResult(object):
    def __init__(self, *, is_done: bool, child_state: Optional['ParserState'] = None) -> None:
        self.is_done = is_done
        self.child_state = child_state


class ParserState(object):
    def tick(self, token: Token, next_token: Token) -> StateTickResult:
        return StateTickResult(is_done=True)

    @property
    def nodes(self) -> Iterable[SyntaxNode]:
        return []


class WhitespaceState(ParserState):
    def tick(self, token: Token, next_token: Token) -> StateTickResult:
        is_done = next_token.type is not TokenType.WHITESPACE
        return StateTickResult(is_done=is_done)


class ExpressionState(ParserState):
    def __init__(self):
        self.args: List[ArgumentNode] = []
        self.arg_parts: List[ArgumentPartNode] = []
        self.is_inside_quotes = False

    def tick(self, token: Token, next_token: Token) -> StateTickResult:
        if token.type is TokenType.WHITESPACE:
            if self.is_inside_quotes:
                token: SymbolToken = token
                # TODO: store whitespace value with whitespace token
                part_node = ArgumentPartNode(ArgumentPartType.CONSTANT, ' ')
                self.arg_parts.append(part_node)
            else:
                self.finish_arg()
            return StateTickResult(is_done=False)

        if token.type is TokenType.EOS:
            if self.is_inside_quotes:
                raise ParseError('Unexpected end of string. Expecting end of quotes.')
            else:
                self.finish_arg()
            return StateTickResult(is_done=True)

        if token.type is TokenType.QUOTES:
            self.is_inside_quotes = not self.is_inside_quotes
            return StateTickResult(is_done=False)

        if token.type is TokenType.SYMBOL:
            token: SymbolToken = token
            part_node = ArgumentPartNode(ArgumentPartType.CONSTANT, token.value)
            self.arg_parts.append(part_node)
            return StateTickResult(is_done=False)

        if token.type is TokenType.REPLACEMENT:
            token: ReplacementToken = token
            type = ArgumentPartType.REPLACEMENT_SINGLE if self.is_inside_quotes else ArgumentPartType.REPLACEMENT
            part_node = ArgumentPartNode(type, token.value)
            self.arg_parts.append(part_node)
            return StateTickResult(is_done=False)

        raise ParseError('Unexpected token while parsing expression: ' + str(token.type))

    def finish_arg(self) -> None:
        if len(self.arg_parts) > 0:
            arg_node = ArgumentNode()
            arg_node.parts = self.arg_parts
            self.arg_parts = []
            self.args.append(arg_node)

    @property
    def nodes(self) -> Iterable[SyntaxNode]:
        return [ExpressionNode(self.args)]


class Parser(object):
    def __init__(self) -> None:
        self.syntax: List[SyntaxNode] = []
        self.tokens: List[Token] = []
        self.state: Optional[ParserState] = None
        self.state_stack: List[ParserState] = []
        self.nodes: List[SyntaxNode] = []

    def parse(self, tokens: List[Token]) -> List[SyntaxNode]:
        self.nodes = []
        self.tokens = tokens

        while len(self.tokens) > 0:
            self.process_token(self.pop_token())

        return self.nodes

    def process_token(self, token: Token) -> None:
        if self.state is None:
            self.state = self.get_next_state(token)

        result = self.state.tick(token, self.peek_token())
        if result.child_state is not None:
            if not result.is_done:
                self.state_stack.append(self.state)
        if result.is_done:
            if len(self.state_stack) > 0:
                self.state = self.state_stack.pop()
            else:
                self.nodes.extend(self.state.nodes)
                self.state = None

    def get_next_state(self, token: Token) -> ParserState:
        type = token.type
        if type == TokenType.WHITESPACE or type == TokenType.EOS:
            return WhitespaceState()
        elif type == TokenType.SYMBOL or type == TokenType.REPLACEMENT:
            return ExpressionState()

        raise ParseError('Unexpected token ' + str(type))

    def peek_token(self) -> Token:
        if len(self.tokens) > 0:
            return self.tokens[0]
        return Token()

    def pop_token(self) -> Token:
        # TODO: This sucks performance wise
        token = self.tokens[0]
        self.tokens = self.tokens[1:]
        return token

    @property
    def is_done(self) -> bool:
        return self.state is None and len(self.state_stack) is 0



import enum
from typing import List


class SyntaxNode(object):
    def accept(self, visitor: 'SyntaxNodeVisitor') -> None:
        pass


class ArgumentPartType(enum.Enum):
    CONSTANT = 0
    REPLACEMENT = 1
    REPLACEMENT_SINGLE = 2


class ArgumentPartNode(SyntaxNode):
    def __init__(self, type: ArgumentPartType, value: str) -> None:
        self.type = type
        self.value = value

    def __repr__(self) -> str:
        return '<ArgumentPartNode type: {0}, value: {1}>'.format(repr(self.type), repr(self.value))

    def accept(self, visitor: 'SyntaxNodeVisitor') -> None:
        visitor.visit_argument_part_node(self)


class ArgumentNode(SyntaxNode):
    def __init__(self) -> None:
        self.parts: List[ArgumentPartNode] = []

    def __repr__(self) -> str:
        return '<ArgumentNode parts: {0}>'.format(repr(self.parts))

    def accept(self, visitor: 'SyntaxNodeVisitor') -> None:
        visitor.visit_argument_node(self)


class ExpressionNode(SyntaxNode):
    def __init__(self, args: List[ArgumentNode]) -> None:
        self.args = args

    def __repr__(self) -> str:
        return '<ExpressionNode args: {0}>'.format(repr(self.args))

    def accept(self, visitor: 'SyntaxNodeVisitor') -> None:
        visitor.visit_expression_node(self)


class SyntaxNodeVisitor(object):
    def visit_argument_part_node(self, node: ArgumentPartNode) -> None:
        pass

    def visit_argument_node(self, node: ArgumentNode) -> None:
        pass

    def visit_expression_node(self, node: ExpressionNode) -> None:
        pass



import sys
from typing import List, Iterable, Dict

from pysh.codegen import Instruction, InstructionVisitor, SubstituteInstruction, CallInstruction, \
    ConcatInstruction, SubstituteSingleInstruction, LoadBufferInstruction, PushBufferInstruction, ResetAInstruction, \
    IncrementAInstruction, PushAInstruction, BranchReturnValueInstruction, BranchBufferEmptyInstruction, \
    JumpRelativeInstruction


class Context(object):
    def __init__(self) -> None:
        self.variables: Dict[str, str] = {}


class ExecutionError(Exception):
    pass


class Interpreter(InstructionVisitor):
    def __init__(self) -> None:
        self.context = Context()
        self.stack: List[str] = []
        self.code: List[Instruction] = []
        self.pc = 0
        self.buffer = ''
        self.reg_a = 0

    def execute(self, code: Iterable[Instruction]) -> None:
        self.code.extend(code)
        code_len = len(self.code)
        try:
            while self.pc < code_len:
                instruction = self.code[self.pc]
                instruction.accept(self)
                self.pc += 1
        except ExecutionError as e:
            sys.stderr.write(str(e))

    def visit_concat(self, instruction: ConcatInstruction) -> None:
        self.buffer += instruction.value

    def visit_substitute(self, instruction: SubstituteInstruction) -> None:
        value = self.context.variables.get(instruction.value, '')
        parts = value.split(' ')

        # remove empty parts
        args = [part for part in parts if len(part) > 0]
        arglen = len(args)

        for i in range(arglen - 1):
            self.buffer += args[i]
            self.stack.append(self.buffer)
            self.buffer = ''

        if arglen > 0:
            self.buffer += args[arglen - 1]
            self.reg_a += arglen - 1

    def visit_substitute_single(self, instruction: SubstituteSingleInstruction) -> None:
        self.buffer += self.context.variables.get(instruction.value, '')

    def visit_load_buffer(self, instruction: LoadBufferInstruction) -> None:
        self.buffer = instruction.value

    def visit_push_buffer(self, instruction: PushBufferInstruction) -> None:
        self.stack.append(self.buffer)
        self.buffer = ''

    def visit_reset_a(self, instruction: ResetAInstruction) -> None:
        self.reg_a = 0

    def visit_increment_a(self, instruction: IncrementAInstruction) -> None:
        self.reg_a += 1

    def visit_push_a(self, instruction: PushAInstruction) -> None:
        self.stack.append(str(self.reg_a))

    def visit_call(self, instruction: CallInstruction) -> None:
        stack_len = len(self.stack)
        if stack_len < self.reg_a:
            raise ExecutionError('Stack underflow! Bad code given to interpreter or interpreter bug.')
        stack_start = stack_len - self.reg_a
        args = self.stack[stack_start:]
        del self.stack[stack_start:]
        print('CALL ' + repr(args))

    def visit_branch_return_value(self, instruction: BranchReturnValueInstruction) -> None:
        raise Exception("TODO")

    def visit_branch_buffer_empty(self, instruction: BranchBufferEmptyInstruction) -> None:
        if len(self.buffer) is 0:
            self.pc += instruction.offset

    def visit_jump_relative(self, instruction: JumpRelativeInstruction) -> None:
        self.pc += instruction.offset



import argparse
import enum
import sys
from typing import Optional, List, Iterable

from pysh.codegen import CodeGenerator, Instruction
from pysh.il import GenerateILVisitor
from pysh.interpreter import Interpreter
from pysh.lexer import Lexer
from pysh.parser import Parser, ParseError
from pysh.syntaxnodes import SyntaxNode


class InteractiveMode(enum.Enum):
    Execute = 0
    GenerateCode = 1
    Parse = 2
    Lex = 3


def print_prompt() -> None:
    sys.stdout.write('pysh$ ')
    sys.stdout.flush()


def print_code(code: Iterable[Instruction]) -> None:
    visitor = GenerateILVisitor()
    for instruction in code:
        instruction.accept(visitor)
    print(visitor.make_il())


def make_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('--mode', choices=('execute', 'codegen', 'parse', 'lex'), default='execute')
    return parser


def main() -> int:
    parser = make_argparser()
    args = parser.parse_args()
    mode = InteractiveMode.Execute
    if args.mode == 'execute':
        mode = InteractiveMode.Execute
    elif args.mode == 'codegen':
        mode = InteractiveMode.GenerateCode
    elif args.mode == 'parse':
        mode = InteractiveMode.Parse
    elif args.mode == 'lex':
        mode = InteractiveMode.Lex

    lexer = Lexer()
    parser = Parser()
    generator = CodeGenerator()
    interpreter = Interpreter()

    print_prompt()

    for line in sys.stdin:
        line = line + '\n'

        ast: Optional[List[SyntaxNode]] = None
        code: Optional[List[Instruction]] = None

        tokens = lexer.lex_all(line)

        try:
            ast = parser.parse(tokens)
        except ParseError as e:
            sys.stderr.write(str(e))
            sys.stderr.write('\n')

        if ast is not None:
            code = generator.generate(ast)

        if mode is InteractiveMode.Execute:
            if code is not None:
                interpreter.execute(code)
        elif mode is InteractiveMode.GenerateCode:
            if code is not None:
                print_code(code)
        elif mode is InteractiveMode.Parse:
            if ast is not None:
                print(repr(ast))
        elif mode is InteractiveMode.Lex:
            print(repr(tokens))

        print_prompt()

    return 0



from typing import List

from pysh.codegen import InstructionVisitor, ConcatInstruction, SubstituteInstruction, SubstituteSingleInstruction, \
    LoadBufferInstruction, PushBufferInstruction, ResetAInstruction, IncrementAInstruction, PushAInstruction, \
    CallInstruction, BranchReturnValueInstruction, BranchBufferEmptyInstruction, JumpRelativeInstruction


class GenerateILVisitor(InstructionVisitor):
    def __init__(self) -> None:
        self.parts: List[str] = []

    def visit_concat(self, instruction: ConcatInstruction) -> None:
        self.parts.append('concat "{0}"\n'.format(instruction.value))

    def visit_substitute(self, instruction: SubstituteInstruction) -> None:
        self.parts.append('sub "{0}"\n'.format(instruction.value))

    def visit_substitute_single(self, instruction: SubstituteSingleInstruction) -> None:
        self.parts.append('subs "{0}"\n'.format(instruction.value))

    def visit_load_buffer(self, instruction: LoadBufferInstruction) -> None:
        self.parts.append('ldbuf "{0}"\n'.format(instruction.value))

    def visit_push_buffer(self, instruction: PushBufferInstruction) -> None:
        self.parts.append('pushbuf\n')

    def visit_reset_a(self, instruction: ResetAInstruction) -> None:
        self.parts.append('reseta\n')

    def visit_increment_a(self, instruction: IncrementAInstruction) -> None:
        self.parts.append('inca\n')

    def visit_push_a(self, instruction: PushAInstruction) -> None:
        self.parts.append('pusha\n')

    def visit_call(self, instruction: CallInstruction) -> None:
        self.parts.append('call\n')

    def visit_branch_return_value(self, instruction: BranchReturnValueInstruction) -> None:
        self.parts.append('brv {0}\n'.format(instruction.offset))

    def visit_branch_buffer_empty(self, instruction: BranchBufferEmptyInstruction) -> None:
        self.parts.append('bbe {0}\n'.format(instruction.offset))

    def visit_jump_relative(self, instruction: JumpRelativeInstruction) -> None:
        self.parts.append('jr {0}\n'.format(instruction.offset))

    def make_il(self) -> str:
        return ''.join(self.parts)
